@@ -18,6 +18,7 @@
 
 #include <string.h>
 #include <assert.h>
+//#include <stdio.h>
 
 #define MAXKC (256 / 32)
 #define MAXKB (256 / 8)
@@ -70,7 +71,7 @@ static inline u32 aes_sbox(u32 in)
 	"	dup 	v1.4s, %w[in]		\n"
 	"	movi	v0.16b, #0		\n"
 	"	aese	v0.16b, v1.16b		\n"
-	"	umov	%w[out], v0.4s[0]		\n"
+	"	umov	%w[out], v0.s[0]	\n"
 	: [out] "=r"(ret)
 	: [in] "r"(in)
 	: "v0", "v1"
@@ -86,21 +87,17 @@ int AES_ARM8_KeySetupEnc(u32 rk[/*4*(Nr + 1)*/], const u8 cipherKey[], int keyBi
 	const int keyln32 = keyBits/32;
 	int i;
 	memcpy(rk, cipherKey, keyBits/8);
-	switch (keyBits) {
-		case 128:
-			if (!rounds)
-				rounds = 10;
-			break;
-		case 192:
-			if (!rounds)
-				rounds = 12;
-			break;
-		case 256:
-			if (!rounds)
-				rounds = 14;
-			break;
-		default:
-			return 0;
+	if (!rounds) {
+		switch (keyBits) {
+			case 128:
+				rounds = 10; break;
+			case 192:
+				rounds = 12; break;
+			case 256:
+				rounds = 14; break;
+			default:
+				return 0;
+		}
 	}
 	for (i = 0; i < sizeof(rcon); ++i) {
 		const u32* rki = rk+i*keyln32;
@@ -131,25 +128,28 @@ int AES_ARM8_KeySetupEnc(u32 rk[/*4*(Nr + 1)*/], const u8 cipherKey[], int keyBi
 	return 0;
 }
 
-inline void AES_ARM8_EKey_DKey(const u32* ekey,
-			   u32* dkey,
-			   int rounds)
+inline void AES_ARM8_EKey_DKey(const u32 *ekey, u32* dkey, int rounds)
 {
-	int i;
-	memcpy(dkey, ekey+rounds*4, 16);
-	for (i = 1, rounds--; rounds > 0; i++, rounds--) {
-		asm volatile(
-		"	ld1	{v0.16b}, %1	\n"
-		"	aesimc	v1.16b, v0.16b	\n"
-		"	st1	{v1.16b}, %0	\n"
-		: "=Q"(dkey[i*4])
-		: "Q"(ekey[rounds*4])
-		: "v0", "v1"
-		);
-	}
-	memcpy(dkey+4*i, ekey, 16);
+	asm volatile(
+		"	ld1	{v0.16b}, [%1]		\n"
+		"	sub	%1, %1, #16		\n"
+		"	st1	{v0.16b}, [%0], #16	\n"
+		"1:					\n"
+		"	ld1	{v0.16b}, [%1]		\n"
+		"	aesimc	v0.16b, v0.16b		\n"
+		"	sub	%1, %1, #16		\n"
+		"	subs 	%w2, %w2, #1		\n"
+		"	st1	{v0.16b}, [%0], #16	\n"
+		"	b.pl	1b			\n"
+		"	ld1	{v0.16b}, [%1]		\n"
+		"//	sub	%1, %1, #16		\n"
+		"	st1	{v0.16b}, [%0], #16	\n"
+		: "=r"(dkey), "=r"(ekey), "=r"(rounds), "=m"(*(roundkey(*)[rounds+1])ekey)
+		: "0"(dkey), "1"(ekey+4*rounds), "2"(rounds-2), "m"(*(roundkey(*)[rounds+1])dkey)
+		: "v0");
 }
 
+
 /**
  * Expand the cipher key into the decryption key schedule.
  *
@@ -158,19 +158,58 @@ inline void AES_ARM8_EKey_DKey(const u32* ekey,
 int AES_ARM8_KeySetupDec(u32 rk[/*4*(Nr + 1)*/], const u8 cipherKey[], int keyBits, int rounds)
 {
 	/* expand the cipher key: */
-	int Nr = AES_ARM8_KeySetupEnc((u32*)crypto->xkeys, cipherKey, keyBits, rounds);
-	AES_ARM8_EKey_DKey((u32*)crypto->xkeys, rk, Nr);
+	int Nr = AES_ARM8_KeySetupEnc(crypto->xkeys->data32, cipherKey, keyBits, rounds);
+	AES_ARM8_EKey_DKey(crypto->xkeys->data32, rk, Nr);
 	return Nr;
 }
 
+#define CLR_NEON3			\
+	asm volatile(			\
+	" movi	v0.16b, #0	\n"	\
+	" movi	v1.16b, #0	\n"	\
+	" movi	v2.16b, #0	\n"	\
+	::: "v0", "v1", "v2")
+
+#define CLR_NEON6			\
+	asm volatile(			\
+	" movi	v0.16b, #0	\n"	\
+	" movi	v1.16b, #0	\n"	\
+	" movi	v2.16b, #0	\n"	\
+	" movi	v3.16b, #0	\n"	\
+	" movi	v4.16b, #0	\n"	\
+	" movi	v5.16b, #0	\n"	\
+	::: "v0", "v1", "v2", "v3", "v4", "v5")
+
+#define CLR_NEON10			\
+	asm volatile(			\
+	" movi	v0.16b, #0	\n"	\
+	" movi	v1.16b, #0	\n"	\
+	" movi	v2.16b, #0	\n"	\
+	" movi	v3.16b, #0	\n"	\
+	" movi	v4.16b, #0	\n"	\
+	" movi	v5.16b, #0	\n"	\
+	" movi	v6.16b, #0	\n"	\
+	" movi	v7.16b, #0	\n"	\
+	" movi	v8.16b, #0	\n"	\
+	" movi	v9.16b, #0	\n"	\
+	::: "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9")
+
+#define CLR_NEON10_14			\
+	asm volatile(			\
+	" movi	v10.16b, #0	\n"	\
+	" movi	v11.16b, #0	\n"	\
+	" movi	v12.16b, #0	\n"	\
+	" movi	v13.16b, #0	\n"	\
+	" movi	v14.16b, #0	\n"	\
+	::: "v10", "v11", "v12", "v13", "v14")
+
+
 void AES_ARM8_Encrypt(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8 pt[16], u8 ct[16])
 {
-	u8 *rk = (u8*)rkeys;
-	uint dummy1;
 	asm volatile(
 	"	ld1	{v0.16b}, [%[pt]]	\n"
 	"	ld1	{v1.4s, v2.4s}, [%[rk]], #32	\n"
-	"//	eor	v0.16b, v0.16b, v1.16b	\n"
+	"//	eor	v0.16b, v0.16b, v3.16b	\n"
 	"	subs	%w[nr], %w[nr], #2	\n"
 	".align 4				\n"
 	"1:					\n"
@@ -192,8 +231,9 @@ void AES_ARM8_Encrypt(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8
 	"	eor	v0.16b, v0.16b, v1.16b	\n"
 	"3:					\n"
 	"	st1	{v0.16b}, [%[ct]]	\n"
-	: [rk] "=r" (rk), [nr] "=r" (dummy1)
-	: "0" (rkeys), "1" (Nr), [pt] "r" (pt), [ct] "r" (ct)
+	: [rk] "=r" (rkeys), [nr] "=r" (Nr), "=m" (*(char(*)[16])ct)
+	//: "0" (rkeys), "1" (Nr), [pt] "r" (pt), [ct] "r" (ct), "m" (*(const char(*)[640])rkeys), "m" (*pt)
+	: "0" (rkeys), "1" (Nr), [pt] "r" (pt), [ct] "r" (ct), "m" (*(const char(*)[16*(Nr+1)])rkeys), "m" (*(const char(*)[16])pt)
 	: "v0", "v1", "v2", "cc"
 	);
 	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
@@ -204,8 +244,6 @@ void AES_ARM8_Encrypt(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8
 
 void AES_ARM8_Decrypt(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8 ct[16], u8 pt[16])
 {
-	u8 *rk = (u8*)rkeys;
-	uint dummy1;
 	asm volatile(
 	"	ld1	{v0.16b}, [%[ct]]	\n"
 	"	ld1	{v1.4s, v2.4s}, [%[rk]], #32	\n"
@@ -231,8 +269,8 @@ void AES_ARM8_Decrypt(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8
 	"	eor	v0.16b, v0.16b, v1.16b	\n"
 	"3:					\n"
 	"	st1	{v0.16b}, [%[pt]]	\n"
-	: [rk] "=r" (rk), [nr] "=r" (dummy1)
-	: "0" (rkeys), "1" (Nr), [ct] "r" (ct), [pt] "r" (pt)
+	: [rk] "=r" (rkeys), [nr] "=r" (Nr), "=m" (*pt)
+	: "0" (rkeys), "1" (Nr), [ct] "r" (ct), [pt] "r" (pt), "m" (*(const char(*)[16*(Nr+1)])rkeys), "m" (*ct)
 	: "v0", "v1", "v2", "cc"
 	);
 	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
@@ -241,8 +279,8 @@ void AES_ARM8_Decrypt(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8
 
 void AES_ARM8_Encrypt4(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8 pt[64], u8 ct[64])
 {
-	u8 *rk = (u8*)rkeys;
-	uint dummy1;
+	/* FIXME: We might need to ensure that pt and ct are not pointing to
+	 * the same register -- we saw this on arm32 (armhf/armv7) */
 	asm volatile(
 	"	ld1	{v2.16b-v5.16b}, [%[pt]]	\n"
 	"	ld1	{v0.4s, v1.4s}, [%[rk]], #32	\n"
@@ -251,49 +289,49 @@ void AES_ARM8_Encrypt4(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8
 	".align 4				\n"
 	"1:					\n"
 	"	aese	v2.16b, v0.16b		\n"
-	"	aese	v3.16b, v0.16b		\n"
-	"	aese	v4.16b, v0.16b		\n"
-	"	aese	v5.16b, v0.16b		\n"
 	"	aesmc	v2.16b, v2.16b		\n"
+	"	aese	v3.16b, v0.16b		\n"
 	"	aesmc	v3.16b, v3.16b		\n"
+	"	aese	v4.16b, v0.16b		\n"
 	"	aesmc	v4.16b, v4.16b		\n"
+	"	aese	v5.16b, v0.16b		\n"
 	"	aesmc	v5.16b, v5.16b		\n"
 	"	ld1	{v0.4s}, [%[rk]], #16	\n"
 	"	b.eq	2f			\n"
 	"	subs	%w[nr], %w[nr], #2	\n"
 	"	aese	v2.16b, v1.16b		\n"
-	"	aese	v3.16b, v1.16b		\n"
-	"	aese	v4.16b, v1.16b		\n"
-	"	aese	v5.16b, v1.16b		\n"
 	"	aesmc	v2.16b, v2.16b		\n"
+	"	aese	v3.16b, v1.16b		\n"
 	"	aesmc	v3.16b, v3.16b		\n"
+	"	aese	v4.16b, v1.16b		\n"
 	"	aesmc	v4.16b, v4.16b		\n"
+	"	aese	v5.16b, v1.16b		\n"
 	"	aesmc	v5.16b, v5.16b		\n"
 	"	ld1	{v1.4s}, [%[rk]], #16	\n"
 	"	b.pl	1b			\n"
 	"					\n"
 	"	aese	v2.16b, v0.16b		\n"
-	"	aese	v3.16b, v0.16b		\n"
-	"	aese	v4.16b, v0.16b		\n"
-	"	aese	v5.16b, v0.16b		\n"
 	"	eor	v2.16b, v2.16b, v1.16b	\n"
+	"	aese	v3.16b, v0.16b		\n"
 	"	eor	v3.16b, v3.16b, v1.16b	\n"
+	"	aese	v4.16b, v0.16b		\n"
 	"	eor	v4.16b, v4.16b, v1.16b	\n"
+	"	aese	v5.16b, v0.16b		\n"
 	"	eor	v5.16b, v5.16b, v1.16b	\n"
 	"	b	3f			\n"
 	"2:					\n"
 	"	aese	v2.16b, v1.16b		\n"
-	"	aese	v3.16b, v1.16b		\n"
-	"	aese	v4.16b, v1.16b		\n"
-	"	aese	v5.16b, v1.16b		\n"
 	"	eor	v2.16b, v2.16b, v0.16b	\n"
+	"	aese	v3.16b, v1.16b		\n"
 	"	eor	v3.16b, v3.16b, v0.16b	\n"
+	"	aese	v4.16b, v1.16b		\n"
 	"	eor	v4.16b, v4.16b, v0.16b	\n"
+	"	aese	v5.16b, v1.16b		\n"
 	"	eor	v5.16b, v5.16b, v0.16b	\n"
 	"3:					\n"
 	"	st1	{v2.16b-v5.16b}, [%[ct]]	\n"
-	: [rk] "=r" (rk), [nr] "=r" (dummy1)
-	: "0" (rkeys), "1" (Nr), [pt] "r" (pt), [ct] "r" (ct)
+	: [rk] "=r" (rkeys), [nr] "=r" (Nr), "=m" (*ct)
+	: "0" (rkeys), "1" (Nr), [pt] "r" (pt), [ct] "r" (ct), "m" (*(const char(*)[16*(Nr+1)])rkeys), "m" (*pt)
 	: "v0", "v1", "v2", "v3", "v4", "v5", "cc"
 	);
 	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
@@ -302,8 +340,6 @@ void AES_ARM8_Encrypt4(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8
 
 void AES_ARM8_Decrypt4(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8 ct[64], u8 pt[64])
 {
-	u8 *rk = (u8*)rkeys;
-	uint dummy1;
 	asm volatile(
 	"	ld1	{v2.16b-v5.16b}, [%[ct]]	\n"
 	"	ld1	{v0.4s, v1.4s}, [%[rk]], #32	\n"
@@ -312,70 +348,71 @@ void AES_ARM8_Decrypt4(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8
 	".align 4				\n"
 	"1:					\n"
 	"	aesd	v2.16b, v0.16b		\n"
-	"	aesd	v3.16b, v0.16b		\n"
-	"	aesd	v4.16b, v0.16b		\n"
-	"	aesd	v5.16b, v0.16b		\n"
 	"	aesimc	v2.16b, v2.16b		\n"
+	"	aesd	v3.16b, v0.16b		\n"
 	"	aesimc	v3.16b, v3.16b		\n"
+	"	aesd	v4.16b, v0.16b		\n"
 	"	aesimc	v4.16b, v4.16b		\n"
+	"	aesd	v5.16b, v0.16b		\n"
 	"	aesimc	v5.16b, v5.16b		\n"
 	"	ld1	{v0.4s}, [%[rk]], #16	\n"
 	"	b.eq	2f			\n"
 	"	subs	%w[nr], %w[nr], #2	\n"
 	"	aesd	v2.16b, v1.16b		\n"
-	"	aesd	v3.16b, v1.16b		\n"
-	"	aesd	v4.16b, v1.16b		\n"
-	"	aesd	v5.16b, v1.16b		\n"
 	"	aesimc	v2.16b, v2.16b		\n"
+	"	aesd	v3.16b, v1.16b		\n"
 	"	aesimc	v3.16b, v3.16b		\n"
+	"	aesd	v4.16b, v1.16b		\n"
 	"	aesimc	v4.16b, v4.16b		\n"
+	"	aesd	v5.16b, v1.16b		\n"
 	"	aesimc	v5.16b, v5.16b		\n"
 	"	ld1	{v1.4s}, [%[rk]], #16	\n"
 	"	b.pl	1b			\n"
 	"					\n"
 	"	aesd	v2.16b, v0.16b		\n"
-	"	aesd	v3.16b, v0.16b		\n"
-	"	aesd	v4.16b, v0.16b		\n"
-	"	aesd	v5.16b, v0.16b		\n"
 	"	eor	v2.16b, v2.16b, v1.16b	\n"
+	"	aesd	v3.16b, v0.16b		\n"
 	"	eor	v3.16b, v3.16b, v1.16b	\n"
+	"	aesd	v4.16b, v0.16b		\n"
 	"	eor	v4.16b, v4.16b, v1.16b	\n"
+	"	aesd	v5.16b, v0.16b		\n"
 	"	eor	v5.16b, v5.16b, v1.16b	\n"
 	"	b	3f			\n"
 	"2:					\n"
 	"	aesd	v2.16b, v1.16b		\n"
-	"	aesd	v3.16b, v1.16b		\n"
-	"	aesd	v4.16b, v1.16b		\n"
-	"	aesd	v5.16b, v1.16b		\n"
 	"	eor	v2.16b, v2.16b, v0.16b	\n"
+	"	aesd	v3.16b, v1.16b		\n"
 	"	eor	v3.16b, v3.16b, v0.16b	\n"
+	"	aesd	v4.16b, v1.16b		\n"
 	"	eor	v4.16b, v4.16b, v0.16b	\n"
+	"	aesd	v5.16b, v1.16b		\n"
 	"	eor	v5.16b, v5.16b, v0.16b	\n"
 	"3:					\n"
 	"	st1	{v2.16b-v5.16b}, [%[pt]]	\n"
-	: [rk] "=r" (rk), [nr] "=r" (dummy1)
-	: "0" (rkeys), "1" (Nr), [ct] "r" (ct), [pt] "r" (pt)
+	: [rk] "=r" (rkeys), [nr] "=r" (Nr), "=m" (*pt)
+	: "0" (rkeys), "1" (Nr), [ct] "r" (ct), [pt] "r" (pt), "m" (*rkeys), "m" (*ct)
 	: "v0", "v1", "v2", "v3", "v4", "v5", "cc"
 	);
 	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
 	return;
 }
 
+// Little-Endian 64bit 1,2,3,4
+static const unsigned long long inc1234[] = {0ULL, 1ULL, 0ULL, 2ULL, 0ULL, 3ULL, 0ULL, 4ULL};
+static const unsigned long long inc1[] = {0ULL, 1ULL};
+
 void AES_ARM8_Encrypt_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8 pt[16], u8 ct[16], u8 iv[16])
 {
-	u8 *rk = (u8*)rkeys;
-	uint dummy1;
-	unsigned long long inc1[] = {0ULL, 1ULL};
 	asm volatile(
 	"	ld1	{v2.16b}, [%[iv]]	\n"
-	"	ld1	{v4.2d}, %[inc]		\n"
+	"	//prfm	PLDL1STRM, [%[pt]]	\n"
+	"	ld1	{v3.2d}, %[inc]		\n"
+	"	subs	w7, %w[nr], #2		\n"
+	"	rev64	v4.16b, v2.16b		\n"
 	"	ld1	{v0.4s, v1.4s}, [%[rk]], #32	\n"
-	"	ld1	{v3.16b}, [%[pt]]	\n"
-	"	rev64	v2.16b, v2.16b		\n"
-	"	subs	%w[nr], %w[nr], #2	\n"
-	"	add	v4.2d, v2.2d, v4.2d	\n"
-	"	rev64	v2.16b, v2.16b		\n"
+	"	add	v4.2d, v4.2d, v3.2d	\n"
 	"	rev64	v4.16b, v4.16b		\n"
+	"	ld1	{v3.16b}, [%[pt]]	\n"
 	"	st1	{v4.16b}, [%[iv]]	\n"
 	".align 4				\n"
 	"1:					\n"
@@ -383,9 +420,9 @@ void AES_ARM8_Encrypt_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const
 	"	aesmc	v2.16b, v2.16b		\n"
 	"	ld1	{v0.4s}, [%[rk]], #16	\n"
 	"	b.eq	2f			\n"
-	"	subs	%w[nr], %w[nr], #2	\n"
 	"	aese	v2.16b, v1.16b		\n"
 	"	aesmc	v2.16b, v2.16b		\n"
+	"	subs	w7, w7, #2		\n"
 	"	ld1	{v1.4s}, [%[rk]], #16	\n"
 	"	b.pl	1b			\n"
 	"					\n"
@@ -398,9 +435,10 @@ void AES_ARM8_Encrypt_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const
 	"3:					\n"
 	"	eor	v3.16b, v3.16b, v2.16b	\n"
 	"	st1	{v3.16b}, [%[ct]]	\n"
-	: [rk] "=r" (rk), [nr] "=r" (dummy1)
-	: "0" (rkeys), "1" (Nr), [pt] "r" (pt), [ct] "r" (ct), [iv] "r" (iv), [inc] "Q" (inc1)
-	: "v0", "v1", "v2", "v3", "v4", "cc"
+	: [rk] "=r" (rkeys), "=m" (*ct), "=m" (*iv)
+	: "0" (rkeys), [nr] "r" (Nr), [pt] "r" (pt), [ct] "r" (ct), [iv] "r" (iv),
+	  [inc] "Q" (inc1), "m" (*(const char(*)[16*(Nr+1)])rkeys), "m" (*pt)
+	: "v0", "v1", "v2", "v3", "v4", "w7", "cc"
 	);
 	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
 	return;
@@ -408,68 +446,65 @@ void AES_ARM8_Encrypt_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const
 
 void AES_ARM8_Encrypt4_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8 pt[64], u8 ct[64], u8 iv[16])
 {
-	u8 *rk = (u8*)rkeys;
-	uint dummy1;
-	unsigned long long inc1[] = {0ULL, 1ULL};
 	asm volatile(
 	"	ld1	{v2.16b}, [%[iv]]	\n"
-	"	ld1	{v10.2d}, %[inc]	\n"
+	"	ld1	{v6.2d-v9.2d}, [%[inc]] \n"
+	"	subs	w7, %w[nr], #2		\n"
+	"	//prfm	PLDL1STRM, [%[pt]]	\n"
+	"	rev64	v5.16b, v2.16b		\n"
 	"	ld1	{v0.4s, v1.4s}, [%[rk]], #32	\n"
-	"	ld1	{v6.16b-v9.16b}, [%[pt]]\n"
-	"	rev64	v2.16b, v2.16b		\n"
-	"	subs	%w[nr], %w[nr], #2	\n"
-	"	add	v3.2d, v2.2d, v10.2d	\n"
-	"	add	v4.2d, v3.2d, v10.2d	\n"
-	"	add	v5.2d, v4.2d, v10.2d	\n"
-	"	add	v10.2d, v5.2d, v10.2d	\n"
-	"	rev64	v2.16b, v2.16b		\n"
+	"	add	v9.2d, v5.2d, v9.2d	\n"
+	"	rev64	v9.16b, v9.16b		\n"
+	"	add	v3.2d, v5.2d, v6.2d	\n"
 	"	rev64	v3.16b, v3.16b		\n"
+	"	add	v4.2d, v5.2d, v7.2d	\n"
 	"	rev64	v4.16b, v4.16b		\n"
-	"	rev64	v5.16b, v5.16b		\n"
-	"	rev64	v10.16b, v10.16b	\n"
-	"	st1	{v10.16b}, [%[iv]]	\n"
+	"	add	v5.2d, v5.2d, v8.2d	\n"
 	"	//prfm	PLDL1STRM, [%[pt],#64]	\n"
+	"	rev64	v5.16b, v5.16b		\n"
+	"	st1	{v9.16b}, [%[iv]]	\n"
+	"	ld1	{v6.16b-v9.16b}, [%[pt]]\n"
 	".align 4				\n"
 	"1:					\n"
 	"	aese	v2.16b, v0.16b		\n"
-	"	aese	v3.16b, v0.16b		\n"
-	"	aese	v4.16b, v0.16b		\n"
-	"	aese	v5.16b, v0.16b		\n"
 	"	aesmc	v2.16b, v2.16b		\n"
+	"	aese	v3.16b, v0.16b		\n"
 	"	aesmc	v3.16b, v3.16b		\n"
+	"	aese	v4.16b, v0.16b		\n"
 	"	aesmc	v4.16b, v4.16b		\n"
+	"	aese	v5.16b, v0.16b		\n"
 	"	aesmc	v5.16b, v5.16b		\n"
 	"	ld1	{v0.4s}, [%[rk]], #16	\n"
 	"	b.eq	2f			\n"
-	"	subs	%w[nr], %w[nr], #2	\n"
+	"	subs	w7, w7, #2		\n"
 	"	aese	v2.16b, v1.16b		\n"
-	"	aese	v3.16b, v1.16b		\n"
-	"	aese	v4.16b, v1.16b		\n"
-	"	aese	v5.16b, v1.16b		\n"
 	"	aesmc	v2.16b, v2.16b		\n"
+	"	aese	v3.16b, v1.16b		\n"
 	"	aesmc	v3.16b, v3.16b		\n"
+	"	aese	v4.16b, v1.16b		\n"
 	"	aesmc	v4.16b, v4.16b		\n"
+	"	aese	v5.16b, v1.16b		\n"
 	"	aesmc	v5.16b, v5.16b		\n"
 	"	ld1	{v1.4s}, [%[rk]], #16	\n"
 	"	b.pl	1b			\n"
 	"					\n"
 	"	aese	v2.16b, v0.16b		\n"
-	"	aese	v3.16b, v0.16b		\n"
-	"	aese	v4.16b, v0.16b		\n"
-	"	aese	v5.16b, v0.16b		\n"
 	"	eor	v2.16b, v2.16b, v1.16b	\n"
+	"	aese	v3.16b, v0.16b		\n"
 	"	eor	v3.16b, v3.16b, v1.16b	\n"
+	"	aese	v4.16b, v0.16b		\n"
 	"	eor	v4.16b, v4.16b, v1.16b	\n"
+	"	aese	v5.16b, v0.16b		\n"
 	"	eor	v5.16b, v5.16b, v1.16b	\n"
 	"	b	3f			\n"
 	"2:					\n"
 	"	aese	v2.16b, v1.16b		\n"
-	"	aese	v3.16b, v1.16b		\n"
-	"	aese	v4.16b, v1.16b		\n"
-	"	aese	v5.16b, v1.16b		\n"
 	"	eor	v2.16b, v2.16b, v0.16b	\n"
+	"	aese	v3.16b, v1.16b		\n"
 	"	eor	v3.16b, v3.16b, v0.16b	\n"
+	"	aese	v4.16b, v1.16b		\n"
 	"	eor	v4.16b, v4.16b, v0.16b	\n"
+	"	aese	v5.16b, v1.16b		\n"
 	"	eor	v5.16b, v5.16b, v0.16b	\n"
 	"3:					\n"
 	"	eor	v6.16b, v6.16b, v2.16b	\n"
@@ -477,32 +512,178 @@ void AES_ARM8_Encrypt4_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, cons
 	"	eor	v8.16b, v8.16b, v4.16b	\n"
 	"	eor	v9.16b, v9.16b, v5.16b	\n"
 	"	st1	{v6.16b-v9.16b}, [%[ct]]	\n"
-	: [rk] "=r" (rk), [nr] "=r" (dummy1)
-	: "0" (rkeys), "1" (Nr), [pt] "r" (pt), [ct] "r" (ct), [iv] "r" (iv), [inc] "Q" (inc1)
-	: "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", "v10", "cc"
+	: [rk] "=r" (rkeys), "=m" (*ct), "=m" (*iv)
+	: "0" (rkeys), [nr] "r" (Nr), [pt] "r" (pt), [ct] "r" (ct),
+	  [iv] "r" (iv), [inc] "r" (inc1234), "m" (*inc1234),
+	  "m" (*(const char(*)[16*(Nr+1)])rkeys), "m" (*pt)
+	: "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", "w7", "cc"
 	);
 	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
 	return;
 }
 
+#ifndef NO_ASM_CTR
+int  AES_ARM8_CTR_Crypt(const uchar* rkeys, uint rounds, uchar *ctr, uint pad,
+			const uchar *input, uchar *output, ssize_t len, ssize_t *olen)
+{
+	*olen = len;
+	asm volatile(
+	"	subs	%[len], %[len], #64	\n"
+	"	ld1	{v10.16b}, [%[iv]]	\n"
+	"	rev64	v10.16b, v10.16b	\n"
+	"	ld1	{v11.2d-v14.2d}, [%[inc]]	\n"
+	"	b.mi	9f			\n"
+	"//.align 4				\n"
+	"0:					\n"
+	"	mov	x8, %[rk]		\n"
+	"	subs	w9, %w[nr], #2		\n"
+	"	ld1	{v0.4s, v1.4s}, [x8], #32	\n"
+	"	//prfm	PSTR2KEEP, [%[ct]]	\n"
+	"	//prfm	PLDL1STRM, [%[pt]]	\n"
+	"	rev64	v2.16b, v10.16b		\n"
+	"	add	v3.2d, v10.2d, v11.2d	\n"
+	"	rev64	v3.16b, v3.16b		\n"
+	"	add	v4.2d, v10.2d, v12.2d	\n"
+	"	rev64	v4.16b, v4.16b		\n"
+	"	add	v5.2d, v10.2d, v13.2d	\n"
+	"	rev64	v5.16b, v5.16b		\n"
+	"	add	v10.2d, v10.2d, v14.2d	\n"
+	"	ld1	{v6.16b-v9.16b}, [%[pt]], #64	\n"
+	".align 4				\n"
+	"1:					\n"
+	"	aese	v2.16b, v0.16b		\n"
+	"	aesmc	v2.16b, v2.16b		\n"
+	"	aese	v3.16b, v0.16b		\n"
+	"	aesmc	v3.16b, v3.16b		\n"
+	"	aese	v4.16b, v0.16b		\n"
+	"	aesmc	v4.16b, v4.16b		\n"
+	"	aese	v5.16b, v0.16b		\n"
+	"	aesmc	v5.16b, v5.16b		\n"
+	"	ld1	{v0.4s}, [x8], #16	\n"
+	"	b.eq	2f			\n"
+	"	subs	w9, w9, #2		\n"
+	"	aese	v2.16b, v1.16b		\n"
+	"	aesmc	v2.16b, v2.16b		\n"
+	"	aese	v3.16b, v1.16b		\n"
+	"	aesmc	v3.16b, v3.16b		\n"
+	"	aese	v4.16b, v1.16b		\n"
+	"	aesmc	v4.16b, v4.16b		\n"
+	"	aese	v5.16b, v1.16b		\n"
+	"	aesmc	v5.16b, v5.16b		\n"
+	"	ld1	{v1.4s}, [x8], #16	\n"
+	"	b.pl	1b			\n"
+	"					\n"
+	"	aese	v2.16b, v0.16b		\n"
+	"	eor	v2.16b, v2.16b, v1.16b	\n"
+	"	aese	v3.16b, v0.16b		\n"
+	"	eor	v3.16b, v3.16b, v1.16b	\n"
+	"	aese	v4.16b, v0.16b		\n"
+	"	eor	v4.16b, v4.16b, v1.16b	\n"
+	"	aese	v5.16b, v0.16b		\n"
+	"	eor	v5.16b, v5.16b, v1.16b	\n"
+	"	b	3f			\n"
+	"2:					\n"
+	"	aese	v2.16b, v1.16b		\n"
+	"	eor	v2.16b, v2.16b, v0.16b	\n"
+	"	aese	v3.16b, v1.16b		\n"
+	"	eor	v3.16b, v3.16b, v0.16b	\n"
+	"	aese	v4.16b, v1.16b		\n"
+	"	eor	v4.16b, v4.16b, v0.16b	\n"
+	"	aese	v5.16b, v1.16b		\n"
+	"	eor	v5.16b, v5.16b, v0.16b	\n"
+	"3:					\n"
+	"	subs	%[len], %[len], #64	\n"
+	"	eor	v6.16b, v6.16b, v2.16b	\n"
+	"	eor	v7.16b, v7.16b, v3.16b	\n"
+	"	eor	v8.16b, v8.16b, v4.16b	\n"
+	"	eor	v9.16b, v9.16b, v5.16b	\n"
+	"	st1	{v6.16b-v9.16b}, [%[ct]], #64	\n"
+	"	b.pl	0b			\n"
+	"9:					\n"
+	"	adds	%[len], %[len], #64	\n"
+	"	b.eq	30f			\n"
+	"10:					\n"
+	"	mov	x8, %[rk]		\n"
+	"	subs	w9, %w[nr], 2		\n"
+	"	ld1	{v0.4s, v1.4s}, [x8], #32	\n"
+	"	rev64	v2.16b, v10.16b		\n"
+	"	add	v10.2d, v10.2d, v11.2d	\n"
+	"	ld1	{v6.16b}, [%[pt]], #16	\n"
+	"//.align 4				\n"
+	"11:					\n"
+	"	aese	v2.16b, v0.16b		\n"
+	"	aesmc	v2.16b, v2.16b		\n"
+	"	ld1	{v0.4s}, [x8], #16	\n"
+	"	b.eq	12f			\n"
+	"	subs	w9, w9, #2		\n"
+	"	aese	v2.16b, v1.16b		\n"
+	"	aesmc	v2.16b, v2.16b		\n"
+	"	ld1	{v1.4s}, [x8], #16	\n"
+	"	b.pl	11b			\n"
+	"					\n"
+	"	aese	v2.16b, v0.16b		\n"
+	"	eor	v2.16b, v2.16b, v1.16b	\n"
+	"	b	13f			\n"
+	"12:					\n"
+	"	aese	v2.16b, v1.16b		\n"
+	"	eor	v2.16b, v2.16b, v0.16b	\n"
+	"13:					\n"
+	"	subs	%[len], %[len], #16	\n"
+	"	eor	v6.16b, v6.16b, v2.16b	\n"
+	"	b.mi	20f			\n"
+	"	st1	{v6.16b}, [%[ct]], #16	\n"
+	"	b.eq	30f			\n"
+	"	b	10b			\n"
+	"20:					\n"
+	"//	add	%[len], %[len], #16	\n"
+	"	tbz	%[len], #3, 21f		\n"
+	"	st1	{v6.d}[0], [%[ct]], #8	\n"
+	"	ext	v6.16b, v6.16b, v6.16b, #8 \n"
+	"21:					\n"
+	"	tbz	%[len], #2, 22f		\n"
+	"	st1	{v6.s}[0], [%[ct]], #4	\n"
+	"	ext	v6.16b, v6.16b, v6.16b, #4 \n"
+	"22:					\n"
+	"	tbz	%[len], #1, 23f		\n"
+	"	st1	{v6.h}[0], [%[ct]], #2	\n"
+	"	ext	v6.16b, v6.16b, v6.16b, #2 \n"
+	"23:					\n"
+	"	tbz	%[len], #0, 30f		\n"
+	"	st1	{v6.b}[0], [%[ct]], #1 \n"
+	"//	ext	v6.16b, v6.16b, v6.16b, #1 \n"
+	"30:					\n"
+	"	rev64	v10.16b, v10.16b	\n"
+	"	st1	{v10.16b}, [%[iv]]	\n"
+	: [len] "=r" (len), [pt] "=r" (input), [ct] "=r" (output),
+	  "=m" (*(char(*)[16])ctr), "=m" (*(char(*)[len])output)
+	: "0" (len), "1" (input), "2" (output), [rk] "r" (rkeys),
+	  [nr] "r" (rounds), [iv] "r" (ctr), [inc] "r" (inc1234),
+	  "m" (*(const char(*)[16*(rounds+1)])rkeys), "m" (*(const char(*)[len])input)
+	: "v0", "v1", "v2", "v3", "v4", "v5", "v6", "x8", "w9", "cc",
+	  "v7", "v8", "v9", "v10", "v11", "v12", "v13", "v14"
+	);
+	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
+	CLR_NEON10;
+	CLR_NEON10_14;
+	//return r;
+	return 0;
+}
+#endif
+
 void AES_ARM8_EncryptX2_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8 pt[16], u8 ct[16], u8 iv[16])
 {
-	u8 *rk = (u8*)rkeys;
-	uint dummy1;
 	assert(Nr > 4 && !(Nr%2));
 	uint halfnr = Nr/2;
-	unsigned long long inc1[] = {0ULL, 1ULL};
 	asm volatile(
 	"	ld1	{v2.16b}, [%[iv]]	\n"
 	"	ld1	{v4.2d}, %[inc]		\n"
+	"	rev64	v3.16b, v2.16b		\n"
 	"	ld1	{v0.4s, v1.4s}, [%[rk]], #32	\n"
-	"	ld1	{v3.16b}, [%[pt]]	\n"
-	"	rev64	v2.16b, v2.16b		\n"
+	"	add	v4.2d, v3.2d, v4.2d	\n"
 	"	subs	%w[nr], %w[nr], #2	\n"
-	"	add	v4.2d, v2.2d, v4.2d	\n"
-	"	rev64	v2.16b, v2.16b		\n"
 	"	rev64	v4.16b, v4.16b		\n"
-	"	mov	w5, %w[nr]		\n"
+	"	ld1	{v3.16b}, [%[pt]]	\n"
+	"	mov	w7, %w[nr]		\n"
 	"	st1	{v4.16b}, [%[iv]]	\n"
 	".align 4				\n"
 	"1:					\n"
@@ -523,18 +704,19 @@ void AES_ARM8_EncryptX2_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, con
 	"	aese	v2.16b, v1.16b		\n"
 	"	eor	v2.16b, v2.16b, v0.16b	\n"
 	"3:					\n"
-	"	cmp	w5, #0			\n"
+	"	cmp	w7, #0			\n"
 	"	b.eq	4f			\n"
 	"	ld1	{v0.4s, v1.4s}, [%[rk]], #32	\n"
-	"	mov	%w[nr], w5		\n"
-	"	mov	w5, #0			\n"
+	"	mov	%w[nr], w7		\n"
+	"	mov	w7, #0			\n"
 	"	b 	1b			\n"
 	"4:					\n"
 	"	eor	v3.16b, v3.16b, v2.16b	\n"
 	"	st1	{v3.16b}, [%[ct]]	\n"
-	: [rk] "=r" (rk), [nr] "=r" (dummy1)
-	: "0" (rkeys), "1" (halfnr), [pt] "r" (pt), [ct] "r" (ct), [iv] "r" (iv), [inc] "Q" (inc1)
-	: "v0", "v1", "v2", "v3", "v4", "w5", "cc"
+	: [rk] "=r" (rkeys), [nr] "=r" (halfnr), "=m" (*ct), "=m" (*iv)
+	: "0" (rkeys), "1" (halfnr), [pt] "r" (pt), [ct] "r" (ct), [iv] "r" (iv),
+	  [inc] "Q" (inc1), "m" (*(const char(*)[16*(Nr+1)])rkeys), "m" (*pt)
+	: "v0", "v1", "v2", "v3", "v4", "w7", "cc"
 	);
 	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
 	return;
@@ -542,78 +724,75 @@ void AES_ARM8_EncryptX2_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, con
 
 void AES_ARM8_Encrypt4X2_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, const u8 pt[64], u8 ct[64], u8 iv[16])
 {
-	u8 *rk = (u8*)rkeys;
-	uint dummy1;
 	assert(Nr > 4 && !(Nr%2));
 	uint halfnr = Nr/2;
-	unsigned long long inc1[] = {0ULL, 1ULL};
 	asm volatile(
 	"	ld1	{v2.16b}, [%[iv]]	\n"
-	"	ld1	{v10.2d}, %[inc]	\n"
+	"	ld1	{v6.2d-v9.2d}, [%[inc]]	\n"
+	"	rev64	v5.16b, v2.16b		\n"
 	"	ld1	{v0.4s, v1.4s}, [%[rk]], #32	\n"
-	"	ld1	{v6.16b-v9.16b}, [%[pt]]\n"
-	"	rev64	v2.16b, v2.16b		\n"
+	"	//prfm	PLDL1STRM, [%[pt]]	\n"
 	"	subs	%w[nr], %w[nr], #2	\n"
-	"	add	v3.2d, v2.2d, v10.2d	\n"
-	"	add	v4.2d, v3.2d, v10.2d	\n"
-	"	add	v5.2d, v4.2d, v10.2d	\n"
-	"	add	v10.2d, v5.2d, v10.2d	\n"
-	"	rev64	v2.16b, v2.16b		\n"
+	"	add	v9.2d, v5.2d, v9.2d	\n"
+	"	rev64	v9.16b, v9.16b		\n"
+	"	add	v3.2d, v5.2d, v6.2d	\n"
 	"	rev64	v3.16b, v3.16b		\n"
+	"	add	v4.2d, v5.2d, v7.2d	\n"
 	"	rev64	v4.16b, v4.16b		\n"
+	"	add	v5.2d, v5.2d, v8.2d	\n"
+	"	mov	w7, %w[nr]		\n"
 	"	rev64	v5.16b, v5.16b		\n"
-	"	rev64	v10.16b, v10.16b	\n"
-	"	mov	w5, %w[nr]		\n"
-	"	st1	{v10.16b}, [%[iv]]	\n"
+	"	st1	{v9.16b}, [%[iv]]	\n"
 	"	//prfm	PLDL1STRM, [%[pt],#64]	\n"
+	"	ld1	{v6.16b-v9.16b}, [%[pt]]\n"
 	".align 4				\n"
 	"1:					\n"
 	"	aese	v2.16b, v0.16b		\n"
-	"	aese	v3.16b, v0.16b		\n"
-	"	aese	v4.16b, v0.16b		\n"
-	"	aese	v5.16b, v0.16b		\n"
 	"	aesmc	v2.16b, v2.16b		\n"
+	"	aese	v3.16b, v0.16b		\n"
 	"	aesmc	v3.16b, v3.16b		\n"
+	"	aese	v4.16b, v0.16b		\n"
 	"	aesmc	v4.16b, v4.16b		\n"
+	"	aese	v5.16b, v0.16b		\n"
 	"	aesmc	v5.16b, v5.16b		\n"
 	"	ld1	{v0.4s}, [%[rk]], #16	\n"
 	"	b.eq	2f			\n"
 	"	subs	%w[nr], %w[nr], #2	\n"
 	"	aese	v2.16b, v1.16b		\n"
-	"	aese	v3.16b, v1.16b		\n"
-	"	aese	v4.16b, v1.16b		\n"
-	"	aese	v5.16b, v1.16b		\n"
 	"	aesmc	v2.16b, v2.16b		\n"
+	"	aese	v3.16b, v1.16b		\n"
 	"	aesmc	v3.16b, v3.16b		\n"
+	"	aese	v4.16b, v1.16b		\n"
 	"	aesmc	v4.16b, v4.16b		\n"
+	"	aese	v5.16b, v1.16b		\n"
 	"	aesmc	v5.16b, v5.16b		\n"
 	"	ld1	{v1.4s}, [%[rk]], #16	\n"
 	"	b.pl	1b			\n"
 	"					\n"
 	"	aese	v2.16b, v0.16b		\n"
-	"	aese	v3.16b, v0.16b		\n"
-	"	aese	v4.16b, v0.16b		\n"
-	"	aese	v5.16b, v0.16b		\n"
 	"	eor	v2.16b, v2.16b, v1.16b	\n"
+	"	aese	v3.16b, v0.16b		\n"
 	"	eor	v3.16b, v3.16b, v1.16b	\n"
+	"	aese	v4.16b, v0.16b		\n"
 	"	eor	v4.16b, v4.16b, v1.16b	\n"
+	"	aese	v5.16b, v0.16b		\n"
 	"	eor	v5.16b, v5.16b, v1.16b	\n"
 	"	b	3f			\n"
 	"2:					\n"
 	"	aese	v2.16b, v1.16b		\n"
-	"	aese	v3.16b, v1.16b		\n"
-	"	aese	v4.16b, v1.16b		\n"
-	"	aese	v5.16b, v1.16b		\n"
 	"	eor	v2.16b, v2.16b, v0.16b	\n"
+	"	aese	v3.16b, v1.16b		\n"
 	"	eor	v3.16b, v3.16b, v0.16b	\n"
+	"	aese	v4.16b, v1.16b		\n"
 	"	eor	v4.16b, v4.16b, v0.16b	\n"
+	"	aese	v5.16b, v1.16b		\n"
 	"	eor	v5.16b, v5.16b, v0.16b	\n"
 	"3:					\n"
-	"	cmp	w5, #0			\n"
+	"	cmp	w7, #0			\n"
 	"	b.eq	4f			\n"
 	"	ld1	{v0.4s, v1.4s}, [%[rk]], #32	\n"
-	"	mov	%w[nr], w5		\n"
-	"	mov	w5, #0			\n"
+	"	mov	%w[nr], w7		\n"
+	"	mov	w7, #0			\n"
 	"	b	1b			\n"
 	"4:					\n"
 	"	eor	v6.16b, v6.16b, v2.16b	\n"
@@ -621,9 +800,11 @@ void AES_ARM8_Encrypt4X2_CTR(const u8 *rkeys /*u32 rk[4*(Nr + 1)]*/, uint Nr, co
 	"	eor	v8.16b, v8.16b, v4.16b	\n"
 	"	eor	v9.16b, v9.16b, v5.16b	\n"
 	"	st1	{v6.16b-v9.16b}, [%[ct]]	\n"
-	: [rk] "=r" (rk), [nr] "=r" (dummy1)
-	: "0" (rkeys), "1" (halfnr), [pt] "r" (pt), [ct] "r" (ct), [iv] "r" (iv), [inc] "Q" (inc1)
-	: "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", "v10", "w5", "cc"
+	: [rk] "=r" (rkeys), [nr] "=r" (halfnr), "=m" (*ct), "=m" (*iv)
+	: "0" (rkeys), "1" (halfnr), [pt] "r" (pt), [ct] "r" (ct),
+          [iv] "r" (iv), [inc] "r" (inc1234), "m" (*inc1234),
+	  "m" (*(const char(*)[16*(Nr+1)])rkeys), "m" (*pt)
+	: "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", "w7", "cc"
 	);
 	//printf("%i rounds left, %li rounds\n", Nr, (rkeys-rk)/16);
 	return;
@@ -654,38 +835,6 @@ DECL_KEYSETUP(Dec, 256);
 #define AES_ARM8_Encrypt_BlkX2_CTR  AES_ARM8_EncryptX2_CTR
 #define AES_ARM8_Encrypt_4BlkX2_CTR AES_ARM8_Encrypt4X2_CTR
 
-#define CLR_NEON3				\
-	asm volatile(				\
-	" movi	v0.16b, #0	\n"	\
-	" movi	v1.16b, #0	\n"	\
-	" movi	v2.16b, #0	\n"	\
-	::: "v0", "v1", "v2")
-
-#define CLR_NEON6				\
-	asm volatile(				\
-	" movi	v0.16b, #0	\n"	\
-	" movi	v1.16b, #0	\n"	\
-	" movi	v2.16b, #0	\n"	\
-	" movi	v3.16b, #0	\n"	\
-	" movi	v4.16b, #0	\n"	\
-	" movi	v5.16b, #0	\n"	\
-	::: "v0", "v1", "v2", "v3", "v4", "v5")
-
-#define CLR_NEON11				\
-	asm volatile(				\
-	" movi	v0.16b, #0	\n"	\
-	" movi	v1.16b, #0	\n"	\
-	" movi	v2.16b, #0	\n"	\
-	" movi	v3.16b, #0	\n"	\
-	" movi	v4.16b, #0	\n"	\
-	" movi	v5.16b, #0	\n"	\
-	" movi	v6.16b, #0	\n"	\
-	" movi	v7.16b, #0	\n"	\
-	" movi	v8.16b, #0	\n"	\
-	" movi	v9.16b, #0	\n"	\
-	" movi	v10.16b, #0	\n"	\
-	::: "v0", "v1", "v2", "v3", "v4", "v5", "v6", "v7", "v8", "v9", "v10")
-
 int  AES_ARM8_ECB_Encrypt(const uchar* rkeys, uint rounds, uchar *iv, uint pad, const uchar *in, uchar *out, ssize_t len, ssize_t *olen)
 {
 	int r = AES_Gen_ECB_Enc4(AES_ARM8_Encrypt_4Blk, AES_ARM8_Encrypt_Blk, 
@@ -700,7 +849,7 @@ int  AES_ARM8_ECB_Decrypt(const uchar* rkeys, uint rounds, uchar *iv, uint pad,
 	CLR_NEON6;
 	return r;
 }
-
+#if 0
 int  AES_ARM8_CBC_Encrypt(const uchar* rkeys, uint rounds, uchar *iv, uint pad, const uchar *in, uchar *out, ssize_t len, ssize_t *olen)
 {
 	int r = AES_Gen_CBC_Enc(AES_ARM8_Encrypt_Blk, 
@@ -708,22 +857,221 @@ int  AES_ARM8_CBC_Encrypt(const uchar* rkeys, uint rounds, uchar *iv, uint pad,
 	CLR_NEON3;
 	return r;
 }
-int  AES_ARM8_CBC_Decrypt(const uchar* rkeys, uint rounds, uchar *iv, uint pad, const uchar *in, uchar *out, ssize_t len, ssize_t *olen)
+#else
+
+#define XOR16(x1, x2, xout)	\
+do {				\
+	uint _i;		\
+	for (_i = 0; _i < 16; _i+=sizeof(ulong))				\
+		*(ulong*)(xout+_i) = *(ulong*)(x1+_i) ^ *(ulong*)(x2+_i);	\
+} while(0)
+
+#define FILL_BLK(in, bf, len, pad) 	\
+do {					\
+	uint i;				\
+	uchar by = pad? 16-(len&0x0f) : 0;	\
+	for (i = 0; i < len; ++i)	\
+		bf[i] = in[i];		\
+	for (; i < 16; ++i)		\
+		bf[i] = by;		\
+} while(0)
+
+
+int  AES_ARM8_CBC_Encrypt(const uchar* rkeys, uint rounds,
+		uchar *iv, uint pad,
+		const uchar *input, uchar *output,
+		ssize_t len, ssize_t *olen)
+{
+	*olen = len;
+	asm volatile(
+	"	subs	%[len], %[len], #16	\n"
+	"	ld1	{v3.16b}, [%[iv]]	\n"
+	"	b.mi	10f			\n"
+	"//.align 4				\n"
+	"0:					\n"
+	"	ld1	{v0.16b}, [%[pt]], #16	\n"
+	"	mov	x8, %[rk]		\n"
+	"	eor	v0.16b, v0.16b, v3.16b	\n"
+	"	ld1	{v1.4s, v2.4s}, [x8], #32	\n"
+	"	subs	w9, %w[nr], #2		\n"
+	".align 4				\n"
+	"1:					\n"
+	"	aese	v0.16b, v1.16b		\n"
+	"	aesmc	v0.16b, v0.16b		\n"
+	"	ld1	{v1.4s}, [x8], #16	\n"
+	"	b.eq	2f			\n"
+	"	subs	w9, w9, #2		\n"
+	"	aese	v0.16b, v2.16b		\n"
+	"	aesmc	v0.16b, v0.16b		\n"
+	"	ld1	{v2.4s}, [x8], #16	\n"
+	"	b.pl	1b			\n"
+	"					\n"
+	"	aese	v0.16b, v1.16b		\n"
+	"	eor	v0.16b, v0.16b, v2.16b	\n"
+	"	b	3f			\n"
+	"2:					\n"
+	"	aese	v0.16b, v2.16b		\n"
+	"	eor	v0.16b, v0.16b, v1.16b	\n"
+	"3:					\n"
+	"	subs	%[len], %[len], #16	\n"
+	"	mov	v3.16b, v0.16b		\n"
+	"	st1	{v0.16b}, [%[ct]], #16	\n"
+	"	b.pl	0b			\n"
+	"10:					\n"
+	"	st1	{v3.16b}, [%[iv]]	\n"
+	"	add	%[len], %[len], #16	\n"
+	"	movi	v3.16b, #0		\n"
+	: [len] "=r" (len), [pt] "=r" (input), [ct] "=r" (output),
+	  "=m" (*(char(*)[16])iv), "=m" (*(char(*)[len])output)
+	: "0" (len), "1" (input), "2" (output), [rk] "r" (rkeys),
+	  [nr] "r" (rounds), [iv] "r" (iv),
+	  "m" (*(const char(*)[16*(rounds+1)])rkeys), "m" (*(const char(*)[len])input)
+	: "v0", "v1", "v2", "v3", "x8", "w9", "cc"
+	);
+	//printf("%li bytes left, %li done\n", len, *olen);
+	if (len || pad == PAD_ALWAYS) {
+		uchar *in = crypto->blkbuf2;
+		FILL_BLK(input, in, len, pad);
+		XOR16(iv, in, iv);
+		AES_ARM8_Encrypt_Blk(rkeys, rounds, iv, output);
+		/* Update last IV */
+		memcpy(iv, output, 16);
+		*olen += 16-(len&15);
+		//memset(in, 0, 16);
+		//LFENCE;
+	}
+	CLR_NEON3;
+	return (pad == PAD_ALWAYS || (len&15))? 16-(len&15): 0;
+}
+#endif
+
+#ifndef NO_CBC_DEC4X
+int  AES_ARM8_CBC_Decrypt(const uchar* rkeys, uint rounds, uchar *iv, uint pad,
+		          const uchar *input, uchar *output, ssize_t len, ssize_t *olen)
+{
+	*olen = len;
+	asm volatile(
+	"	subs	%[len], %[len], #64	\n"
+	"	//prfm	PLDL1STRM, [%[ct]]	\n"
+	"	ld1	{v2.16b}, [%[iv]]	\n"
+	"	b.mi	10f			\n"
+	"//.align 4				\n"
+	"0:					\n"
+	"	mov	x8, %[rk]		\n"
+	"	ld1	{v3.16b-v6.16b}, [%[ct]], #64\n"
+	"	//prfm	PLDL1STRM, [%[ct],#64]	\n"
+	"	ld1	{v0.4s, v1.4s}, [x8], #32	\n"
+	"	subs	w9, %w[nr], #2		\n"
+	"	mov	v7.16b, v3.16b		\n"
+	"	mov	v8.16b, v4.16b		\n"
+	"	mov	v9.16b, v5.16b		\n"
+	"	mov	v10.16b, v6.16b		\n"
+	".align 4				\n"
+	"1:					\n"
+	"	aesd	v3.16b, v0.16b		\n"
+	"	aesimc	v3.16b, v3.16b		\n"
+	"	aesd	v4.16b, v0.16b		\n"
+	"	aesimc	v4.16b, v4.16b		\n"
+	"	aesd	v5.16b, v0.16b		\n"
+	"	aesimc	v5.16b, v5.16b		\n"
+	"	aesd	v6.16b, v0.16b		\n"
+	"	aesimc	v6.16b, v6.16b		\n"
+	"	ld1	{v0.4s}, [x8], #16	\n"
+	"	b.eq	2f			\n"
+	"	subs	w9, w9, #2		\n"
+	"	aesd	v3.16b, v1.16b		\n"
+	"	aesimc	v3.16b, v3.16b		\n"
+	"	aesd	v4.16b, v1.16b		\n"
+	"	aesimc	v4.16b, v4.16b		\n"
+	"	aesd	v5.16b, v1.16b		\n"
+	"	aesimc	v5.16b, v5.16b		\n"
+	"	aesd	v6.16b, v1.16b		\n"
+	"	aesimc	v6.16b, v6.16b		\n"
+	"	ld1	{v1.4s}, [x8], #16	\n"
+	"	b.pl	1b			\n"
+	"					\n"
+	"	aesd	v3.16b, v0.16b		\n"
+	"	eor	v3.16b, v3.16b, v1.16b	\n"
+	"	aesd	v4.16b, v0.16b		\n"
+	"	eor	v4.16b, v4.16b, v1.16b	\n"
+	"	aesd	v5.16b, v0.16b		\n"
+	"	eor	v5.16b, v5.16b, v1.16b	\n"
+	"	aesd	v6.16b, v0.16b		\n"
+	"	eor	v6.16b, v6.16b, v1.16b	\n"
+	"	b	3f			\n"
+	"2:					\n"
+	"	aesd	v3.16b, v1.16b		\n"
+	"	eor	v3.16b, v3.16b, v0.16b	\n"
+	"	aesd	v4.16b, v1.16b		\n"
+	"	eor	v4.16b, v4.16b, v0.16b	\n"
+	"	aesd	v5.16b, v1.16b		\n"
+	"	eor	v5.16b, v5.16b, v0.16b	\n"
+	"	aesd	v6.16b, v1.16b		\n"
+	"	eor	v6.16b, v6.16b, v0.16b	\n"
+	"3:					\n"
+	"	subs	%[len], %[len], #64	\n"
+	"	eor	v3.16b, v3.16b, v2.16b	\n"
+	"	eor	v4.16b, v4.16b, v7.16b	\n"
+	"	eor	v5.16b, v5.16b, v8.16b	\n"
+	"	eor	v6.16b, v6.16b, v9.16b	\n"
+	"	mov	v2.16b, v10.16b		\n"
+	"	st1	{v3.16b-v6.16b}, [%[pt]], #64	\n"
+	"	b.pl	0b			\n"
+	"10:					\n"
+	"	st1	{v2.16b}, [%[iv]]	\n"
+	"	add	%[len], %[len], #64	\n"
+	"	//movi	v6.16b, #0		\n"
+	"	movi	v10.16b, #0		\n"
+	: [len] "=r" (len), [ct] "=r" (input), [pt] "=r" (output),
+	  "=m" (*(char(*)[16])iv), "=m" (*(char(*)[len])output)
+	: "0" (len), "1" (input), "2" (output), [rk] "r" (rkeys),
+	  [nr] "r" (rounds), [iv] "r" (iv),
+	  "m" (*(const char(*)[16*(rounds+1)])rkeys), "m" (*(const char(*)[len])input)
+	: "v0", "v1", "v2", "v3", "v4", "v5", "v6", "x8", "w9", "cc",
+	  "v7", "v8", "v9", "v10"
+	);
+	//printf("%li bytes left, %li done\n", len, *olen);
+	while (len > 0) {
+		uchar *ebf = crypto->blkbuf3;
+		AES_ARM8_Decrypt_Blk(rkeys, rounds, input, ebf);
+		XOR16(iv, ebf, output);
+		/* Update last IV */
+		memcpy(iv, input, 16);
+		//memset(in, 0, 16);
+		//LFENCE;
+		len -= 16; input += 16; output += 16;
+	}
+	CLR_NEON10;
+	if (pad)
+		return dec_fix_olen_pad(olen, pad, output);
+	else
+		return 0;
+	//return (pad == PAD_ALWAYS || (len&15))? 16-(len&15): 0;
+}
+
+
+#else
+int  AES_ARM8_CBC_Decrypt(const uchar* rkeys, uint rounds, uchar *iv, uint pad,
+			  const uchar *in, uchar *out, ssize_t len, ssize_t *olen)
 {
 	int r = AES_Gen_CBC_Dec4(AES_ARM8_Decrypt_4Blk, AES_ARM8_Decrypt_Blk, 
 				rkeys, rounds, iv, pad, in, out, len, olen);
 	CLR_NEON6;
 	return r;
 }
+#endif
 
-int  AES_ARM8_CTR_Crypt(const uchar* rkeys, uint rounds, uchar *ctr, uint pad, const uchar *in, uchar *out, ssize_t len, ssize_t *olen)
+#ifdef NO_ASM_CTR
+int  AES_ARM8_CTR_Crypt(const uchar* rkeys, uint rounds, uchar *ctr, uint pad,
+			const uchar *in, uchar *out, ssize_t len, ssize_t *olen)
 {
 	*olen = len;
 	int r = AES_Gen_CTR_Crypt_Opt(AES_ARM8_Encrypt_4Blk_CTR, AES_ARM8_Encrypt_Blk_CTR, 
 				     rkeys, rounds, ctr, in, out, len);
-	CLR_NEON11;
+	CLR_NEON10;
 	return r;
 }
+#endif
 
 /* Double de/encryption methods */
 
@@ -836,7 +1184,7 @@ int  AES_ARM8_CTR_CryptX2(const uchar* rkeys, uint rounds, uchar *ctr, uint pad,
 	//int r = AES_Gen_CTR_Crypt(AES_ARM8_Encrypt_BlkX2, rkeys, rounds, ctr, in, out, len);
 	int r = AES_Gen_CTR_Crypt_Opt(AES_ARM8_Encrypt_4BlkX2_CTR, AES_ARM8_Encrypt_BlkX2_CTR,
 				     rkeys, rounds, ctr, in, out, len);
-	CLR_NEON11;
+	CLR_NEON10;
 	return r;
 }
 
