@@ -7,10 +7,13 @@
  * License: GPLv2 or v3 or BSD (3-clause)
  */
 
-#include <openssl/evp.h>
-#include <assert.h>
 #include "aes.h"
 #include "aes_ossl.h"
+#ifdef HAVE_OPENSSL_EVP_H
+#include <openssl/evp.h>
+#endif
+#include <assert.h>
+#include "secmem.h"
 #include <string.h>
 
 #include <netinet/in.h>
@@ -60,27 +63,28 @@ int AES_OSSL_##BITCHAIN##_Encrypt(const unsigned char* ctx, unsigned int rounds,
 	if (IV) {						\
 		memcpy(evpctx->oiv, iv, 16); memcpy(evpctx->iv, iv, 16);	\
 	}							\
+	if (!len && !pad) { *flen = 0; return 0; }		\
        	if (DOPAD && !pad && (len&15)) {			\
 		ores = EVP_EncryptUpdate(evpctx, out, &olen, in, len-(len&15));	\
 		assert(ores);					\
-		uchar ibf[16];					\
+		uchar *ibf = crypto->blkbuf2;			\
 		memcpy(ibf, in+olen, len&15);			\
 		memset(ibf+(len&15), 0, 16-(len&15));		\
 		ores = EVP_EncryptUpdate(evpctx, out+olen, &elen, ibf, 16);	\
 		memset(ibf, 0, len&15);				\
-		asm("":::"memory");				\
+		LFENCE;						\
 		assert(ores);					\
 	} else {								\
-		if (DOPAD && !(len%15) && pad == PAD_ASNEEDED)	\
+		if (DOPAD && !(len%16) && pad == PAD_ASNEEDED)	\
 			EVP_CIPHER_CTX_set_padding(evpctx, 0);	\
 		ores = EVP_EncryptUpdate(evpctx, out, &olen, in, len);		\
 		assert(ores);					\
-		ores = EVP_EncryptFinal(evpctx, out+olen, &elen);\
+		ores = EVP_EncryptFinal(evpctx, out+olen, &elen);		\
 		assert(ores);					\
 		if (0 && elen && (len&15)) olen -= 16;		\
 	}							\
 	*flen = olen+elen;					\
-	if (DOPAD && pad == PAD_ASNEEDED && !(len&15))		\
+	if (0 && DOPAD && pad == PAD_ASNEEDED && !(len&15))			\
 		*flen -= 16;					\
 	if (0 && olen+elen < len)				\
 		fprintf(stderr, "Encryption length mismatch %i+%i != %zi\n",	\
@@ -102,9 +106,10 @@ int AES_OSSL_##BITCHAIN##_Decrypt(const unsigned char* ctx, unsigned int rounds,
 	if (IV) {						\
 		memcpy(evpctx->oiv, iv, 16); memcpy(evpctx->iv, iv, 16);	\
 	}							\
+	if (!len && pad != PAD_ALWAYS) { *flen = 0; return 0; }	\
 	if (DOPAD && pad == PAD_ASNEEDED) {			\
 		int olen1;					\
-		uchar buf[16];					\
+		uchar *buf = crypto->blkbuf3;			\
 		ores = EVP_DecryptUpdate(evpctx, out, &olen, in, ilen-16);	\
 		assert(ores);					\
 		EVP_CIPHER_CTX ctx2;				\
@@ -127,7 +132,7 @@ int AES_OSSL_##BITCHAIN##_Decrypt(const unsigned char* ctx, unsigned int rounds,
 			assert(ores);				\
 		}						\
 		memset(&ctx2, 0, sizeof(ctx2));			\
-		asm("":::"memory");				\
+		LFENCE;						\
 	} else {						\
 		ores = EVP_DecryptUpdate(evpctx, out, &olen, in, ilen);	\
 		assert(ores);					\
@@ -140,7 +145,7 @@ int AES_OSSL_##BITCHAIN##_Decrypt(const unsigned char* ctx, unsigned int rounds,
 	if (IV)							\
 		memcpy(iv, evpctx->iv, 16);			\
 	if (DOPAD && pad == PAD_ASNEEDED)			\
-		return (elen? 16-elen: 1);			\
+		return (elen? 16-elen: ILLEGAL_PADDING);	\
 	return ores - 1;					\
 }
 
@@ -201,7 +206,7 @@ void AES_OSSL_Bits_EKey_ExpandX2(const EVP_CIPHER *cipher, const unsigned char*
 	EVP_EncryptInit_ex(evpctx+1, cipher, NULL, usrkey2, NULL);
 	//EVP_CIPHER_CTX_set_padding(evpctx+1, 0);
 	memset(usrkey2, 0, 32);
-	asm("":::"memory");
+	LFENCE;
 }
 void AES_OSSL_Bits_DKey_ExpandX2(const EVP_CIPHER *cipher, const unsigned char* userkey, unsigned char *ctx, unsigned int bits)
 {
@@ -219,7 +224,7 @@ void AES_OSSL_Bits_DKey_ExpandX2(const EVP_CIPHER *cipher, const unsigned char*
 	EVP_DecryptInit_ex(evpctx+1, cipher, NULL, usrkey2, NULL);
 	//EVP_CIPHER_CTX_set_padding(evpctx+1, 0);
 	memset(usrkey2, 0, 32);
-	asm("":::"memory");
+	LFENCE;
 }
 
 void AES_OSSL_RecycleX2(unsigned char *ctx)
@@ -259,15 +264,16 @@ int  AES_OSSL_##BITCHAIN##_EncryptX2(const unsigned char* ctx, unsigned int roun
 		memcpy(evpctx->oiv, iv, 16); memcpy(evpctx->iv, iv, 16);		\
 		memcpy((evpctx+1)->oiv, iv, 16); memcpy((evpctx+1)->iv, iv, 16);	\
 	}							\
+	if (!len && !pad) { *flen = 0; return 0; }		\
        	if (!pad && (len&15)) {					\
 		ores = EVP_EncryptUpdate(evpctx, out, &olen, in, len-(len&15));		\
 		assert(ores);					\
-		uchar ibf[16];					\
+		uchar *ibf = crypto->blkbuf2;			\
 		memcpy(ibf, in+olen, len&15);			\
 		memset(ibf+(len&15), 0, 16-(len&15));		\
 		ores = EVP_EncryptUpdate(evpctx, out+olen, &elen, ibf, 16);		\
 		memset(ibf, 0, len&15);				\
-		asm("":::"memory");				\
+		LFENCE;						\
 		assert(ores);					\
 	} else {						\
 		ores = EVP_EncryptUpdate(evpctx, out, &olen, in, len);	\
@@ -300,13 +306,14 @@ int  AES_OSSL_##BITCHAIN##_DecryptX2(const unsigned char* ctx, unsigned int roun
 		memcpy((evpctx+1)->oiv, iv, 16); memcpy((evpctx+1)->iv, iv, 16);	\
 		memcpy(evpctx->oiv, iv, 16); memcpy(evpctx->iv, iv, 16);		\
 	}							\
+	if (!len && pad != PAD_ALWAYS) { *flen = 0; return 0; }	\
 	ores = EVP_DecryptUpdate(evpctx+1, out, &olen, in, rlen);		\
 	assert(ores);						\
 	ores = EVP_DecryptFinal(evpctx+1, out+olen, &elen);	\
 	assert(ores);						\
 	if (pad == PAD_ASNEEDED) {				\
 		int ilen = olen, olen1;				\
-		uchar buf[16];					\
+		uchar *buf = crypto->blkbuf3;			\
 		ores = EVP_DecryptUpdate(evpctx, out, &olen, out, ilen-16);	\
 		assert(ores); assert(olen == ilen-16);		\
 		/* Save piece that gets overwritten */		\
@@ -327,7 +334,7 @@ int  AES_OSSL_##BITCHAIN##_DecryptX2(const unsigned char* ctx, unsigned int roun
 			assert(ores);				\
 		}						\
 		memset(&ctx2, 0, sizeof(ctx2));			\
-		asm("":::"memory");				\
+		LFENCE;						\
 	} else {						\
 		ores = EVP_DecryptUpdate(evpctx, out, &olen, out, olen+elen);	\
 		assert(ores);						\
@@ -340,7 +347,7 @@ int  AES_OSSL_##BITCHAIN##_DecryptX2(const unsigned char* ctx, unsigned int roun
 	if (IV)							\
 		memcpy(iv, evpctx->iv, 16);			\
 	if (pad == PAD_ASNEEDED)				\
-		return (elen? 16-elen: 1);			\
+		return (elen? 16-elen: ILLEGAL_PADDING);	\
 	return ores - 1;					\
 }
 
@@ -380,22 +387,22 @@ void AES_OSSL_Blk_EncryptX2(const unsigned char *ctx, unsigned int rounds,
 {
 	EVP_CIPHER_CTX *evpctx = (EVP_CIPHER_CTX*)ctx;
 	int olen;
-	uchar blk[16];
+	uchar *blk = crypto->blkbuf1;
 	EVP_EncryptUpdate(evpctx, blk, &olen, in, 16);
 	EVP_EncryptUpdate(evpctx+1, out, &olen, blk, olen);
 	memset(blk, 0, 16);
-	asm("":::"memory");
+	LFENCE;
 }
 void AES_OSSL_Blk_DecryptX2(const unsigned char *ctx, unsigned int rounds,
 			    const unsigned char *in, unsigned char *out)			
 {
 	EVP_CIPHER_CTX *evpctx = (EVP_CIPHER_CTX*)ctx;
 	int olen;
-	uchar blk[16];
+	uchar *blk = crypto->blkbuf1;
 	EVP_DecryptUpdate(evpctx+1, blk, &olen, in, 16);
 	EVP_DecryptUpdate(evpctx, out, &olen, blk, olen);
 	memset(blk, 0, 16);
-	asm("":::"memory");
+	LFENCE;
 }
 
 
